unit SIP;

// SIP Client for Delphi 6PE, (c)2006 by Paul TOTH <tothpaul@free.fr>
// http://tothpaul.free.fr

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

(* Documentation :
 ---------------

 Session Initiation Protocol (RFC 3261) :

  http://fr.wikipedia.org/wiki/Session_Initiation_Protocol
  ftp://ftp.rfc-editor.org/in-notes/rfc3261.txt

 Session Description Protocol (RFC 2327) :

 Other links :

  http://abcdrfc.free.fr/ [RFC en franais]


 Historique :
 ------------

 11 aot 2006
  - correction d'un petit bug dans Crunch()

 5 aot 2006
  - codage du mot de passe, le calcul A1 peut remplacer le mot de passe sous la forme {A1}

 1er aot 2006
 - premire release des sources

 17 juillet 2006 - b0.1.2 :
  - rcriture partie UDP

 12 juillet 2006 - b0.1.1 :
  - passage en mode asynchrone
  - avec gestion de l'tat de la "communication" avec OnStep
  - gestion de l'interactivit (SIP.Cancel)

  NB: il faut revoir tout l'aspect socket I/O car en UDP a ne devrait pas fonctionner comme je l'ai fait

 11 juillet 2006 - b0.1 :
  - premiers tests en mode bloquant, a fonctionne en UDP, mais il reste la partie RTP/AVP !

 SDP (RFC 2327) 

*)

interface

uses
 Windows, Messages, Winsock,
 Classes, SysUtils;

const
 SIP_RELEASE  = 'SIPInside 0.2.3 (2006-08-11)';

 SIP_DEFAULT_PORT = 5060;
 
// SIP Error
 SIP_NO_ERROR        = 0;
 SIP_HOST_NOT_FOUND  = 1;
 SIP_NO_SOCKET       = 2;
 SIP_SEND_ERROR      = 3;
 SIP_RECV_ERROR      = 4;
 SIP_INVALID_RESPONSE= 5;
 SIP_UNAUTHORIZED    = 6;
 SIP_NOT_CONNECTED   = 7;

type
 TCallStep=(
  csNONE,     // before or after a conversation
  csINVITE,   // send INVITE
  csAUTH,     // 40x Authenticate
  csTRYING,   // 101 Trying
  csWRONG,    // 403 Wrong login or password
  csRINGING,  // 108 Ringing
  csCANCEL,   // send CANCEL
  csCANCELED, // 487 Canceled, or cancel INVITE (no response from server)
  csONLINE,   // 200 OK after INVITE, TRYING or RINGING
  csSENDBYE,  // send BYE
  csRECVBYE   // BYE received
 );

 TCustomSIPClient=class
 private
  fHWnd     :THandle;   // Window Handle used for Async sockets
  fMsg      :integer;   // Message for Async sockets
  fName     :string;    // User name
 private
  fUserID   :string;    // SIP userid
  fPasswd   :string;    // SIP password
  fPhone    :string;    // Phone number to call -> <sip:fPhone@fDomain>
  fSession  :integer;   // SDP session
 // Events
  fError    :integer;   // Last error number (OnError)
 // Socket
  fSocket   :integer;   // UDP socket
  fRemote   :TSockAddr; // SockAddr to send data to fDomain
  fReceived :string;    // fDomain IP address
  fVia      :string;    // SIP Via header line
  fContact  :string;    // SIP Contact header line
  fBranch   :string;    // SIP Via header line Branch value
  fFrom     :string;    // SIP From header line
  fCallID   :string;    // SIP Call-ID
  fRequest  :string;    // last Request
  fURI      :string;    // URI
  fTarget   :string;    // Target
  fTagged   :string;    // Tagged target
  fSequence :cardinal;  // Request sequence
  fStep     :TCallStep; // Phone call step
 // socket I/O
  procedure ReadData;
  procedure WriteData;
 // raising error
  procedure SetError(Value:integer);
  procedure FatalError(ErrNo:integer);
 // extract header info
  function IntegerAt(Offset:integer):integer;
  function getHeader(const Tag:string):string;
 // generic send request method
  procedure SendRequest(const Request,Target,Headers,Content:string);
 // send a request with a new Request number
  procedure ExecRequest(const Request,Target,Headers,Content:string);
 // retreive SIP status code
  function GetStatus:integer;
 // send a status code for a SIP request (BYE)
  procedure SendReply(const Status:string);
 // compute authorisation response (RFC 2617)
  function GetAuthorization(const Method,URI:string):string;
 // resend request with authorization response
  procedure SendAuthorization;
 // not used
  procedure SendRegister;
 protected
  fDomain   :string;    // SIP domain
  fData     :string;    // request body
  fSend     :string;    // send buffer
  fHead     :string;    // request header
  fStatus   :integer;   // request status code
  fLocalSDP :string;
  fRemoteSDP:string;
  procedure RecvData; virtual;
  procedure SendData; virtual;
  procedure Start; virtual;
  procedure Close; virtual;
  procedure SetStep(Value:TCallStep); virtual;
  property Error:integer read fError write SetError;
  property Step:TCallStep read fStep write SetStep;
 public
  constructor Create(AHandle:HWnd; AMsg:integer);
  destructor Destroy; override;
  procedure SetLocal(const Addr:string; Port:word);
  procedure Call(const AUsername, ADomain, AUserID, APassword, APhoneNumber:string);
  function Crunch(const UserID, Realm, Password:string):string;
  procedure Cancel;
  procedure ProcessMessage(var Msg);
 end;

 TSIPClient=class(TCustomSIPClient)
 private
  EOnSend :TNotifyEvent;
  EOnRecv :TNotifyEvent;
  EOnStep :TNotifyEvent;
  EOnClose:TNotifyEvent;
  EOnError:TNotifyEvent;
  EOnStart:TNotifyEvent;
 protected
  procedure RecvData; override;
  procedure SendData; override;
  procedure Start; override;
  procedure SetStep(Value:TCallStep); override;
 public
  procedure Close; override;
  property Domain:string read fDomain;
  property Error;
  property Step;
  property Status:integer read fStatus;
  property Head:string read fHead;
  property Data:string read fData;
  property Send:string read fSend;
  property LocalSDP:string read fLocalSDP write fLocalSDP;
  property RemoteSDP:string read fRemoteSDP;
  property OnSend:TNotifyEvent read EOnSend write EOnSend;
  property OnRecv:TNotifyEvent read EOnRecv write EOnRecv;
  property OnStep:TNotifyEvent read EOnStep write EOnStep;
  property OnClose:TNotifyEvent read EOnClose write EOnClose;
  property OnStart:TNotifyEvent read EOnStart write EOnStart;
  property OnError:TNotifyEvent read EOnError write EOnError;
 end;

implementation

uses md5, UDP;

function getValue(const key,str:string):string;
var
 i,j:integer;
begin
 Result:='';
 i:=pos(key,str);
 if i=0 then exit;
 inc(i,length(key));
 if i+2>length(str) then exit;
 if str[i]<>'=' then exit;
 inc(i);
 if str[i]='"' then begin
  inc(i);
  j:=i;
  while (j<=length(str)) and (str[j]<>'"') do inc(j);
 end else begin
  j:=i;
  while (j<=length(str))and(not (str[j] in [' ',','])) do inc(j);
 end;
 Result:=copy(str,i,j-i);
end;

//----------------------------------------------------------------------------//

constructor TCustomSIPClient.Create(AHandle:HWnd; AMsg:integer);
begin
 inherited Create;
 fHwnd  :=AHandle;
 fMsg   :=AMsg;
 fCallID:=CreateID;
 fSocket:=INVALID_SOCKET;
 fSession:=getTickCount;
end;

destructor TCustomSIPClient.Destroy;
begin
 Close;
end;

procedure TCustomSIPClient.SetLocal(const Addr:string; Port:word);
begin
 fLocalSDP:='v=0'#13#10+
            'o=SIPInside '+IntToStr(fSession)+' '+IntToStr(getTickCount)+' IN IP4 '+Addr+#13#10+
            's=SIP Call'#13#10+
            'c=IN IP4 '+Addr+#13#10+
            't=0 0'#13#10+
            'm=audio '+IntToStr(Port)+' RTP/AVP 8'#13#10+
            'b=AS:64'#13#10+
            'a=rtpmap:8 PCMA/8000/1'#13#10+
            'a=ptime:30'#13#10;
end;

procedure TCustomSIPClient.Call(const AUsername, ADomain, AUserID, APassword, APhoneNumber:string);
var
 ad:integer;
begin
 Close;
// Parms
 fName  :=AUsername;
 fDomain:=ADomain;
 fUserID:=AUserID;
 fPasswd:=APassword;
 fPhone :=APhoneNumber;
// Socket
 fHead:='';
 fData:='';
// find host address
 ad:=INetAddr(fDomain);
 if ad=INADDR_NONE then FatalError(SIP_HOST_NOT_FOUND);
 fReceived:=inet_ntoa(in_addr(ad));
 // open a socket
 fSocket:=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
 if fSocket=INVALID_SOCKET then FatalError(SIP_NO_SOCKET);
 // ASyncrhone
 Error:=WSAASyncSelect(fSocket,fHWnd,fMsg,FD_READ or FD_WRITE);
 // call server
 FillChar(fRemote,SizeOf(fRemote),0);
 fRemote.sin_family:=AF_INET;
 fRemote.sin_port:=htons(SIP_DEFAULT_PORT);
 Error:=bind(fSocket,fRemote,SizeOf(fRemote));
 fRemote.sin_addr.S_addr:=ad;
// Setup request header
 fVia    :='Via: SIP/2.0/UDP '+fDomain+';rport;branch=z9hG4bK';
 fFrom   :='From: '+fName+' <sip:'+fUserID+'@'+fDomain+'>;tag='+IntToStr(GetTickCount)+#13#10;
 fContact:='Contact: '+fName+' <sip:'+fUserID+'@'+fDomain+'>'#13#10;
// Call the remote number
 fURI   :='sip:'+fPhone+'@'+fDomain;
 fTarget:='<'+fURI+'>';
 fTagged:='';
 fStep:=csNONE;
 Step:=csINVITE;
 ExecRequest('INVITE',fTarget,'Content-Type: application/sdp'#13#10,fLocalSDP);
end;

procedure TCustomSIPClient.Cancel;
begin
 case fStep of
  csINVITE : begin
   Step:=csCANCELED;
   Close;
  end;
  csAUTH, csTRYING, csRINGING : begin
   Step:=csCANCEL;
  // use SendRequest to keep the same Request number
   SendRequest('CANCEL',fTarget,'','');
  end;
  csONLINE : begin
   Step:=csSENDBYE;
  // use ExecRequest to provide a new Request number
   ExecRequest('BYE',fTagged,'','');
  end;
  else Close;
 end;
end;

procedure TCustomSIPClient.Close;
begin
 if fSocket=INVALID_SOCKET then exit;
 WSAASyncSelect(fSocket,fHWnd,0,0);
 closesocket(fSocket);
 fSocket:=INVALID_SOCKET;
end;

procedure TCustomSIPClient.ProcessMessage(var Msg);
begin
 with TMessage(Msg) do begin
  case lParam of
   FD_READ   : ReadData;
   FD_WRITE  : WriteData;
   else        Close;
  end;
 end;
end;

procedure TCustomSIPClient.ReadData;
var
 l,i :integer;
 head:integer;
 len :integer;
begin
 if fSocket=INVALID_SOCKET then exit;
 fHead:='';
 SetLength(fData,1024);
 l:=recv(fSocket,fData[1],1024,0);
 if l=0 then FatalError(SIP_RECV_ERROR);
 if l<0 then begin
  i:=WSAGetLastError;
  if i=WSAEWOULDBLOCK then exit;
  FatalError(i);
 end;
 // Cirpack KeepAlive Packet ?!
 if copy(fData,1,24)='Cirpack KeepAlive Packet' then exit;
// get Header first
 head:=pos(#13#10#13#10,fData);
 if head=0 then exit; // need more data
 fHead:=copy(fData,1,head+1);
 i:=pos('Content-Length: ',fHead);
 if i=0 then FatalError(SIP_INVALID_RESPONSE);
 len:=IntegerAt(i+16);
// now content
 if l<head+len then exit; // need more data
 fData:=copy(fData,head+4,len);
 RecvData;
end;

procedure TCustomSIPClient.WriteData;
var
 e,len:integer;
begin
 if fSocket=INVALID_SOCKET then exit;
 len:=Length(fSend);
 while len>0 do begin
  e:=sendto(fSocket,fSend[1],len,0,fRemote,SizeOf(fRemote));
  if e=0 then FatalError(SIP_SEND_ERROR);
  if e<0 then begin
   e:=WSAGetLastError;
   if e=WSAEWOULDBLOCK then break;
   FatalError(e);
  end;
  if e=len then begin
   len:=0;
   fSend:='';
  end;
 end;
end;

procedure TCustomSIPClient.RecvData;
begin
// BYE from the remote client ?
 if copy(fHead,1,4)='BYE ' then begin
  //if fStep=csONLINE then begin
   SendReply('200 OK');
   Step:=csRECVBYE;
   Close;
  //end;
 end else begin
  fTagged:=getHeader('To');
  if fTagged='' then fTagged:=fTarget;
  case GetStatus of
  // 100 Trying
   100: if fStep in [csINVITE,csAUTH] then Step:=csTRYING;
  // 180 Ringing
   180: if fStep in [csINVITE,csAUTH,csTRYING] then Step:=csRINGING;
  // 200 OK
   200: case fStep of
         csINVITE,
         csAUTH,
         csTRYING,
         csRINGING: begin
          Step:=csONLINE;
          Start;
          SendRequest('ACK',fTagged,'','');
         end;
         csSENDBYE: begin
          Step:=csNONE;
          Close;
         end;
        end;
  // 403 Wrong login or password
   403: begin
         Step:=csWRONG;
         SendRequest('ACK',fTagged,'',''); // WRONG
         Close;
        end;
  // 401 WWW-Authenticate
  // 407 Proxy-Authenticate
   401,
   407: if fStep=csINVITE then begin
         Step:=csAUTH;
         SendAuthorization;
        end;
  // 487 Canceled
   487: if fStep=csCANCEL then begin
         Step:=csCANCELED;
         SendRequest('ACK',fTagged,'',''); // CANCELED
         Close;
        end;
  end;
 end;
 ReadData;
end;

procedure TCustomSIPClient.Start;
begin
// fData = remote SDP
 fRemoteSDP:=fData;
end;

procedure TCustomSIPClient.SendData;
begin
 if fSocket=INVALID_SOCKET then FatalError(SIP_NOT_CONNECTED);
 WriteData;
end;

procedure TCustomSIPClient.SetError(Value:integer);
begin
 if Value=SOCKET_ERROR then begin
  fError:=WSAGetLastError;
  if (fError=0)or(fError=WSAEWOULDBLOCK) then exit;
  FatalError(fError);
 end;
end;

procedure TCustomSIPClient.FatalError(ErrNo:integer);
begin
 fError:=ErrNo;
 Abort;
end;

function TCustomSIPClient.IntegerAt(Offset:integer):integer;
var
 num:char;
begin
 Result:=0;
 while (Offset<=Length(fHead)) do begin
  num:=fHead[Offset];
  if num in ['0'..'9'] then
   Result:=10*Result+ord(num)-ord('0')
  else
   exit;
  inc(Offset);
 end;
end;

function TCustomSIPClient.getHeader(const Tag:string):string;
var
 i,j:integer;
begin
 Result:='';
 i:=pos(#13#10+Tag+':',fHead);
 if i=0 then exit;
 inc(i,2+Length(Tag)+1);
 j:=i;
 while (j<=Length(fHead))and(fHead[j]<>#13) do inc(j);
 Result:=copy(fHead,i,j-i);
end;

procedure TCustomSIPClient.SendRequest(const Request,Target,Headers,Content:string);
begin
 fRequest:=Request;
 fSend:=Request+' '+fURI+' SIP/2.0'#13#10+
        fVia+fBranch+#13#10+
        fFrom+
        'To: '+Target+#13#10+
        'Call-ID: '+fCallID+'@'+fDomain+#13#10+
        'CSeq: '+IntToStr(fSequence)+' '+Request+#13#10+
        'Max-Forwards: 70'#13#10+
        Headers+
        'Content-Length: '+IntToStr(Length(Content))+#13#10+
        #13#10+
        Content;
 SendData;
end;

function TCustomSIPClient.GetStatus:integer;
begin
 if (copy(fHead,1,8)<>'SIP/2.0 ') then FatalError(SIP_INVALID_RESPONSE);
 fStatus:=IntegerAt(9);
 if (fStatus<100)or(fStatus>699)  then FatalError(SIP_INVALID_RESPONSE);
 Result:=fStatus;
end;

procedure TCustomSIPClient.SendReply(const Status:string);
begin
{
 fData:='SIP/2.0 '+Status+copy(fHead,pos(#13#10,fHead),MaxInt);
}
 fSend:='SIP/2.0 '+Status+#13#10+
        'Via:'+getHeader('Via')+';received='+fReceived+#13#10+
        'From:'+getHeader('From')+#13#10+
        'To:'+getHeader('To')+#13#10+
        'Call-ID:'+getHeader('Call-ID')+#13#10+
        'CSeq:'+getHeader('CSeq')+#13#10+
        'Content-Length: 0'+#13#10+
        #13#10;
 SendData;
end;

procedure TCustomSIPClient.ExecRequest(const Request,Target,Headers,Content:string);
begin
 fBranch:=CreateID;
 inc(fSequence);
 SendRequest(
  Request,Target,
  fContact+
 'Expires: 1800'#13#10+
 'User-Agent: '+SIP_RELEASE+#13#10+
  Headers,
  Content
 );
 ReadData;
end;

procedure TCustomSIPClient.SendRegister;
begin
 fURI:='sip:'+fDomain;
 ExecRequest('REGISTER',fName+' <sip:'+fUserID+'@'+fDomain+'>','','');
end;

procedure TCustomSIPClient.SendAuthorization;
var
 reqs:string;
 auth:string;
begin
 reqs:=fRequest;
 auth:=GetAuthorization(fRequest,fURI);
 SendRequest('ACK',fTagged,'','');
 if reqs='INVITE' then
  ExecRequest(reqs,fTarget,auth+'Content-Type: application/sdp'#13#10,fLocalSDP)
 else
  ExecRequest(reqs,fTarget,auth,'');
end;

function TCustomSIPClient.Crunch(const UserID, Realm, Password:string):string;
begin
 if (Length(Password)=2+32)and(Password[1]='{')and(Password[34]='}') then
  Result:=copy(Password,2,32)
 else
  Result:=MD5Print(MD5String(UserID+':'+Realm+':'+Password));
end;

function TCustomSIPClient.GetAuthorization(const Method,URI:string):string; // RFC 2617
var
 auth  :string;
 realm :string;
 nonce :string;
 opaque:string;
 A1,A2 :string;
 resp  :string;
begin
 Result:='';
 if fStatus=407 then begin
  auth:=getHeader('Proxy-Authenticate');
  if auth='' then FatalError(SIP_UNAUTHORIZED);
  Result:='Proxy-';
 end else begin
  auth:=getHeader('WWW-Authenticate');
  if auth='' then FatalError(SIP_UNAUTHORIZED);
 end;

 if getValue('algorithm',auth)<>'MD5' then FatalError(SIP_UNAUTHORIZED);

 realm :=getValue('realm' ,auth);
 nonce :=getValue('nonce' ,auth);
 opaque:=getValue('opaque',auth);

 A1:=Crunch(fUserID,realm,fPasswd);
 A2:=MD5Print(MD5String(Method+':'+URI));
 resp:=MD5Print(MD5String(A1+':'+nonce+':'+A2));

 Result:=Result+'Authorization: Digest username="'+fUserID+'",realm="'+realm+'",nonce="'+nonce+'",response="'+resp+'",uri="'+URI+'",algorithm=MD5,opaque="'+opaque+'"'#13#10;
end;

procedure TCustomSIPClient.SetStep(Value:TCallStep);
begin
 fStep:=Value;
end;

//----------------------------------------------------------------------------//
procedure TSIPClient.RecvData;
begin
 if Assigned(EOnRecv) then EOnRecv(Self);
 inherited;
end;

procedure TSIPClient.SendData;
begin
 if Assigned(EOnSend) then EOnSend(Self);
 inherited;
end;

procedure TSIPClient.Start;
begin
 inherited;
 if Assigned(EOnStart) then EOnStart(Self);
end;

procedure TSIPClient.SetStep(Value:TCallStep);
begin
 inherited;
 if Assigned(EOnStep) then EOnStep(Self)
end;

procedure TSIPClient.Close;
begin
 if (fSocket<>INVALID_SOCKET) and Assigned(EOnClose) then EOnClose(Self);
 inherited;
end;

//----------------------------------------------------------------------------//

procedure InitSocket;
var
 wsa:TWSAData;
begin
 WSAStartup($101,wsa);
end;

initialization
 InitSocket;
end.

